@@ -7,9 +7,10 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  */
 
-#include "ChildSession.hpp"
 #include "config.h"
 
+#include "ChildSession.hpp"
+
 #include <sstream>
 
 #include <Poco/JSON/Object.h>
@@ -37,7 +38,7 @@ std::recursive_mutex ChildSession::Mutex;
 ChildSession::ChildSession(const std::string& id,
                            const std::string& jailId,
                            IDocumentManager& docManager) :
-    LOOLSession(id, Kind::ToMaster, nullptr),
+    Session("ToMaster-" + id, id, false),
     _jailId(jailId),
     _docManager(docManager),
     _viewId(-1),
@@ -68,14 +69,15 @@ void ChildSession::disconnect()
             LOG_WRN("Skipping unload on incomplete view.");
         }
 
-        LOOLSession::disconnect();
+        Session::disconnect();
     }
 }
 
 bool ChildSession::_handleInput(const char *buffer, int length)
 {
+    LOG_TRC(getName() + ": handling [" << getAbbreviatedMessage(buffer, length) << "].");
     const std::string firstLine = getFirstLine(buffer, length);
-    StringTokenizer tokens(firstLine, " ", StringTokenizer::TOK_IGNORE_EMPTY | StringTokenizer::TOK_TRIM);
+    const auto tokens = LOOLProtocol::tokenize(firstLine.data(), firstLine.size());
 
     if (LOOLProtocol::tokenIndicatesUserInteraction(tokens[0]))
     {
@@ -83,7 +85,7 @@ bool ChildSession::_handleInput(const char *buffer, int length)
         updateLastActivityTime();
     }
 
-    if (tokens.count() > 0 && tokens[0] == "useractive" && getLOKitDocument() != nullptr)
+    if (tokens.size() > 0 && tokens[0] == "useractive" && getLOKitDocument() != nullptr)
     {
         LOG_DBG("Handling message after inactivity of " << getInactivityMS() << "ms.");
         setIsActive(true);
@@ -95,19 +97,14 @@ bool ChildSession::_handleInput(const char *buffer, int length)
 
         getLOKitDocument()->setView(_viewId);
 
-        // Get the list of view ids from the core
-        const int viewCount = getLOKitDocument()->getViewsCount();
-        std::vector<int> viewIds(viewCount);
-        getLOKitDocument()->getViewIds(viewIds.data(), viewCount);
-
         int curPart = 0;
         if (getLOKitDocument()->getDocumentType() != LOK_DOCTYPE_TEXT)
             curPart = getLOKitDocument()->getPart();
 
-        lockLokDoc.unlock();
-
         // Notify all views about updated view info
-        _docManager.notifyViewInfo(viewIds);
+        _docManager.notifyViewInfo();
+
+        lockLokDoc.unlock();
 
         if (getLOKitDocument()->getDocumentType() != LOK_DOCTYPE_TEXT)
         {
@@ -115,20 +112,47 @@ bool ChildSession::_handleInput(const char *buffer, int length)
             sendTextFrame("setpart: part=" + std::to_string(curPart));
         }
 
-        //TODO: Is the order of these important?
-        for (const auto& pair : _lastDocEvents)
+        // Invalidate if we have to
+        // TODO instead just a "_invalidate" flag, we should remember / grow
+        // the rectangle to invalidate; invalidating everything is sub-optimal
+        if (_stateRecorder._invalidate)
         {
-            const auto typeName = LOKitHelper::kitCallbackTypeToString(pair.first);
-            LOG_TRC("Replaying missed event: " << typeName << ": " << pair.second);
-            loKitCallback(pair.first, pair.second);
+            std::string payload = "0, 0, " + std::to_string(INT_MAX) + ", " + std::to_string(INT_MAX) + ", " + std::to_string(curPart);
+            loKitCallback(LOK_CALLBACK_INVALIDATE_TILES, payload);
         }
 
-        for (const auto& pair : _lastDocStates)
+        for (const auto& viewPair : _stateRecorder._recordedViewEvents)
         {
-            LOG_TRC("Replaying missed state-change: STATE_CHANGED: " << pair.second);
+            for (const auto& eventPair : viewPair.second)
+            {
+                const RecordedEvent& event = eventPair.second;
+                LOG_TRC("Replaying missed view event: " <<  viewPair.first << " " << LOKitHelper::kitCallbackTypeToString(event._type)
+                                                        << ": " << event._payload);
+                loKitCallback(event._type, event._payload);
+            }
+        }
+
+        for (const auto& eventPair : _stateRecorder._recordedEvents)
+        {
+            const RecordedEvent& event = eventPair.second;
+            LOG_TRC("Replaying missed event: " << LOKitHelper::kitCallbackTypeToString(event._type) << ": " << event._payload);
+            loKitCallback(event._type, event._payload);
+        }
+
+        for (const auto& pair : _stateRecorder._recordedStates)
+        {
+            LOG_TRC("Replaying missed state-change: " << pair.second);
             loKitCallback(LOK_CALLBACK_STATE_CHANGED, pair.second);
         }
 
+        for (const auto& event : _stateRecorder._recordedEventsVector)
+        {
+            LOG_TRC("Replaying missed event (part of sequence): " << LOKitHelper::kitCallbackTypeToString(event._type) << ": " << event._payload);
+            loKitCallback(event._type, event._payload);
+        }
+
+        _stateRecorder.clear();
+
         LOG_TRC("Finished replaying messages.");
     }
 
@@ -141,10 +165,6 @@ bool ChildSession::_handleInput(const char *buffer, int length)
     {
         return getCommandValues(buffer, length, tokens);
     }
-    else if (tokens[0] == "partpagerectangles")
-    {
-        return getPartPageRectangles(buffer, length);
-    }
     else if (tokens[0] == "load")
     {
         if (_isDocLoaded)
@@ -288,10 +308,10 @@ bool ChildSession::_handleInput(const char *buffer, int length)
     return true;
 }
 
-bool ChildSession::loadDocument(const char * /*buffer*/, int /*length*/, StringTokenizer& tokens)
+bool ChildSession::loadDocument(const char * /*buffer*/, int /*length*/, const std::vector<std::string>& tokens)
 {
     int part = -1;
-    if (tokens.count() < 2)
+    if (tokens.size() < 2)
     {
         sendTextFrame("error: cmd=load kind=syntax");
         return false;
@@ -316,7 +336,7 @@ bool ChildSession::loadDocument(const char * /*buffer*/, int /*length*/, StringT
 
     std::unique_lock<std::recursive_mutex> lock(Mutex);
 
-    bool loaded = _docManager.onLoad(getId(), _jailedFilePath, _userName, _docPassword, renderOpts, _haveDocPassword);
+    bool loaded = _docManager.onLoad(getId(), _jailedFilePath, _userName, _docPassword, renderOpts, _haveDocPassword, _lang);
     if (!loaded || _viewId < 0)
     {
         LOG_ERR("Failed to get LoKitDocument instance.");
@@ -345,25 +365,19 @@ bool ChildSession::loadDocument(const char * /*buffer*/, int /*length*/, StringT
         return false;
     }
 
-    // Get the list of view ids from the core
-    const int viewCount = getLOKitDocument()->getViewsCount();
-    std::vector<int> viewIds(viewCount);
-    getLOKitDocument()->getViewIds(viewIds.data(), viewCount);
-
-    lockLokDoc.unlock();
-
     // Inform everyone (including this one) about updated view info
-    _docManager.notifyViewInfo(viewIds);
+    _docManager.notifyViewInfo();
 
     LOG_INF("Loaded session " << getId());
     return true;
 }
 
-bool ChildSession::sendFontRendering(const char* /*buffer*/, int /*length*/, StringTokenizer& tokens)
+bool ChildSession::sendFontRendering(const char* /*buffer*/, int /*length*/, const std::vector<std::string>& tokens)
 {
-    std::string font, text, decodedFont;
+    std::string font, text, decodedFont, decodedChar;
+    bool bSuccess;
 
-    if (tokens.count() < 3 ||
+    if (tokens.size() < 3 ||
         !getTokenString(tokens[1], "font", font))
     {
         sendTextFrame("error: cmd=renderfont kind=syntax");
@@ -372,8 +386,19 @@ bool ChildSession::sendFontRendering(const char* /*buffer*/, int /*length*/, Str
 
     getTokenString(tokens[2], "char", text);
 
-    URI::decode(font, decodedFont);
-    std::string response = "renderfont: " + Poco::cat(std::string(" "), tokens.begin() + 1, tokens.end()) + "\n";
+    try
+    {
+        URI::decode(font, decodedFont);
+        URI::decode(text, decodedChar);
+    }
+    catch (Poco::SyntaxException& exc)
+    {
+        LOG_DBG(exc.message());
+        sendTextFrame("error: cmd=renderfont kind=syntax");
+        return false;
+    }
+
+    const std::string response = "renderfont: " + Poco::cat(std::string(" "), tokens.begin() + 1, tokens.end()) + "\n";
 
     std::vector<char> output;
     output.resize(response.size());
@@ -388,20 +413,27 @@ bool ChildSession::sendFontRendering(const char* /*buffer*/, int /*length*/, Str
 
         getLOKitDocument()->setView(_viewId);
 
-        ptrFont = getLOKitDocument()->renderFont(decodedFont.c_str(), text.c_str(), &width, &height);
+        ptrFont = getLOKitDocument()->renderFont(decodedFont.c_str(), decodedChar.c_str(), &width, &height);
     }
 
     LOG_TRC("renderFont [" << font << "] rendered in " << (timestamp.elapsed()/1000.) << "ms");
 
-    if (!ptrFont ||
-        !png::encodeBufferToPNG(ptrFont, width, height, output, LOK_TILEMODE_RGBA))
+    if (!ptrFont)
     {
-        std::free(ptrFont);
-        return sendTextFrame("error: cmd=renderfont kind=failure");
+        return sendTextFrame(output.data(), output.size());
+    }
+
+    if (Png::encodeBufferToPNG(ptrFont, width, height, output, LOK_TILEMODE_RGBA))
+    {
+        bSuccess = sendTextFrame(output.data(), output.size());
+    }
+    else
+    {
+        bSuccess = sendTextFrame("error: cmd=renderfont kind=failure");
     }
 
     std::free(ptrFont);
-    return sendTextFrame(output.data(), output.size());
+    return bSuccess;
 }
 
 bool ChildSession::getStatus(const char* /*buffer*/, int /*length*/)
@@ -456,12 +488,12 @@ void insertUserNames(const std::map<int, UserInfo>& viewInfo, std::string& json)
 
 }
 
-bool ChildSession::getCommandValues(const char* /*buffer*/, int /*length*/, StringTokenizer& tokens)
+bool ChildSession::getCommandValues(const char* /*buffer*/, int /*length*/, const std::vector<std::string>& tokens)
 {
     bool success;
-    char* pValues;
+    char* values;
     std::string command;
-    if (tokens.count() != 2 || !getTokenString(tokens[1], "command", command))
+    if (tokens.size() != 2 || !getTokenString(tokens[1], "command", command))
     {
         sendTextFrame("error: cmd=commandvalues kind=syntax");
         return false;
@@ -473,43 +505,35 @@ bool ChildSession::getCommandValues(const char* /*buffer*/, int /*length*/, Stri
 
     if (command == ".uno:DocumentRepair")
     {
-        char* pUndo;
+        char* undo;
         const std::string jsonTemplate("{\"commandName\":\".uno:DocumentRepair\",\"Redo\":%s,\"Undo\":%s}");
-        pValues = getLOKitDocument()->getCommandValues(".uno:Redo");
-        pUndo = getLOKitDocument()->getCommandValues(".uno:Undo");
+        values = getLOKitDocument()->getCommandValues(".uno:Redo");
+        undo = getLOKitDocument()->getCommandValues(".uno:Undo");
         std::string json = Poco::format(jsonTemplate,
-                                        std::string(pValues == nullptr ? "" : pValues),
-                                        std::string(pUndo == nullptr ? "" : pUndo));
+                                        std::string(values == nullptr ? "" : values),
+                                        std::string(undo == nullptr ? "" : undo));
         // json only contains view IDs, insert matching user names.
         std::map<int, UserInfo> viewInfo = _docManager.getViewInfo();
         insertUserNames(viewInfo, json);
         success = sendTextFrame("commandvalues: " + json);
-        std::free(pValues);
-        std::free(pUndo);
+        std::free(values);
+        std::free(undo);
     }
     else
     {
-        pValues = getLOKitDocument()->getCommandValues(command.c_str());
-        success = sendTextFrame("commandvalues: " + std::string(pValues == nullptr ? "" : pValues));
-        std::free(pValues);
+        values = getLOKitDocument()->getCommandValues(command.c_str());
+        success = sendTextFrame("commandvalues: " + std::string(values == nullptr ? "" : values));
+        std::free(values);
     }
 
     return success;
 }
 
-bool ChildSession::getPartPageRectangles(const char* /*buffer*/, int /*length*/)
-{
-    // We don't support partpagerectangles any more, will be removed in the
-    // next version
-    sendTextFrame("partpagerectangles: ");
-    return true;
-}
-
-bool ChildSession::clientZoom(const char* /*buffer*/, int /*length*/, StringTokenizer& tokens)
+bool ChildSession::clientZoom(const char* /*buffer*/, int /*length*/, const std::vector<std::string>& tokens)
 {
     int tilePixelWidth, tilePixelHeight, tileTwipWidth, tileTwipHeight;
 
-    if (tokens.count() != 5 ||
+    if (tokens.size() != 5 ||
         !getTokenInteger(tokens[1], "tilepixelwidth", tilePixelWidth) ||
         !getTokenInteger(tokens[2], "tilepixelheight", tilePixelHeight) ||
         !getTokenInteger(tokens[3], "tiletwipwidth", tileTwipWidth) ||
@@ -527,14 +551,14 @@ bool ChildSession::clientZoom(const char* /*buffer*/, int /*length*/, StringToke
     return true;
 }
 
-bool ChildSession::clientVisibleArea(const char* /*buffer*/, int /*length*/, StringTokenizer& tokens)
+bool ChildSession::clientVisibleArea(const char* /*buffer*/, int /*length*/, const std::vector<std::string>& tokens)
 {
     int x;
     int y;
     int width;
     int height;
 
-    if (tokens.count() != 5 ||
+    if (tokens.size() != 5 ||
         !getTokenInteger(tokens[1], "x", x) ||
         !getTokenInteger(tokens[2], "y", y) ||
         !getTokenInteger(tokens[3], "width", width) ||
@@ -552,11 +576,11 @@ bool ChildSession::clientVisibleArea(const char* /*buffer*/, int /*length*/, Str
     return true;
 }
 
-bool ChildSession::downloadAs(const char* /*buffer*/, int /*length*/, StringTokenizer& tokens)
+bool ChildSession::downloadAs(const char* /*buffer*/, int /*length*/, const std::vector<std::string>& tokens)
 {
     std::string name, id, format, filterOptions;
 
-    if (tokens.count() < 5 ||
+    if (tokens.size() < 5 ||
         !getTokenString(tokens[1], "name", name) ||
         !getTokenString(tokens[2], "id", id))
     {
@@ -568,7 +592,7 @@ bool ChildSession::downloadAs(const char* /*buffer*/, int /*length*/, StringToke
 
     if (getTokenString(tokens[4], "options", filterOptions))
     {
-        if (tokens.count() > 5)
+        if (tokens.size() > 5)
         {
             filterOptions += Poco::cat(std::string(" "), tokens.begin() + 5, tokens.end());
         }
@@ -600,11 +624,11 @@ bool ChildSession::getChildId()
     return true;
 }
 
-bool ChildSession::getTextSelection(const char* /*buffer*/, int /*length*/, StringTokenizer& tokens)
+bool ChildSession::getTextSelection(const char* /*buffer*/, int /*length*/, const std::vector<std::string>& tokens)
 {
     std::string mimeType;
 
-    if (tokens.count() != 2 ||
+    if (tokens.size() != 2 ||
         !getTokenString(tokens[1], "mimetype", mimeType))
     {
         sendTextFrame("error: cmd=gettextselection kind=syntax");
@@ -625,11 +649,11 @@ bool ChildSession::getTextSelection(const char* /*buffer*/, int /*length*/, Stri
     return true;
 }
 
-bool ChildSession::paste(const char* buffer, int length, StringTokenizer& tokens)
+bool ChildSession::paste(const char* buffer, int length, const std::vector<std::string>& tokens)
 {
     std::string mimeType;
-
-    if (tokens.count() < 2 || !getTokenString(tokens[1], "mimetype", mimeType))
+    if (tokens.size() < 2 || !getTokenString(tokens[1], "mimetype", mimeType) ||
+        mimeType.empty())
     {
         sendTextFrame("error: cmd=paste kind=syntax");
         return false;
@@ -637,21 +661,23 @@ bool ChildSession::paste(const char* buffer, int length, StringTokenizer& tokens
 
     const std::string firstLine = getFirstLine(buffer, length);
     const char* data = buffer + firstLine.size() + 1;
-    const size_t size = length - firstLine.size() - 1;
+    const int size = length - firstLine.size() - 1;
+    if (size > 0)
+    {
+        std::unique_lock<std::mutex> lock(_docManager.getDocumentMutex());
 
-    std::unique_lock<std::mutex> lock(_docManager.getDocumentMutex());
+        getLOKitDocument()->setView(_viewId);
 
-    getLOKitDocument()->setView(_viewId);
-
-    getLOKitDocument()->paste(mimeType.c_str(), data, size);
+        getLOKitDocument()->paste(mimeType.c_str(), data, size);
+    }
 
     return true;
 }
 
-bool ChildSession::insertFile(const char* /*buffer*/, int /*length*/, StringTokenizer& tokens)
+bool ChildSession::insertFile(const char* /*buffer*/, int /*length*/, const std::vector<std::string>& tokens)
 {
     std::string name, type;
-    if (tokens.count() != 3 ||
+    if (tokens.size() != 3 ||
         !getTokenString(tokens[1], "name", name) ||
         !getTokenString(tokens[2], "type", type))
     {
@@ -679,10 +705,10 @@ bool ChildSession::insertFile(const char* /*buffer*/, int /*length*/, StringToke
     return true;
 }
 
-bool ChildSession::keyEvent(const char* /*buffer*/, int /*length*/, StringTokenizer& tokens)
+bool ChildSession::keyEvent(const char* /*buffer*/, int /*length*/, const std::vector<std::string>& tokens)
 {
     int type, charcode, keycode;
-    if (tokens.count() != 4 ||
+    if (tokens.size() != 4 ||
         !getTokenKeyword(tokens[1], "type",
                          {{"input", LOK_KEYEVENT_KEYINPUT}, {"up", LOK_KEYEVENT_KEYUP}},
                          type) ||
@@ -718,7 +744,7 @@ bool ChildSession::keyEvent(const char* /*buffer*/, int /*length*/, StringTokeni
     return true;
 }
 
-bool ChildSession::mouseEvent(const char* /*buffer*/, int /*length*/, StringTokenizer& tokens)
+bool ChildSession::mouseEvent(const char* /*buffer*/, int /*length*/, const std::vector<std::string>& tokens)
 {
     int type, x, y, count;
     bool success = true;
@@ -727,7 +753,7 @@ bool ChildSession::mouseEvent(const char* /*buffer*/, int /*length*/, StringToke
     int buttons = 1; // left button
     int modifier = 0;
 
-    if (tokens.count() < 5 ||
+    if (tokens.size() < 5 ||
         !getTokenKeyword(tokens[1], "type",
                          {{"buttondown", LOK_MOUSEEVENT_MOUSEBUTTONDOWN},
                           {"buttonup", LOK_MOUSEEVENT_MOUSEBUTTONUP},
@@ -741,11 +767,11 @@ bool ChildSession::mouseEvent(const char* /*buffer*/, int /*length*/, StringToke
     }
 
     // compatibility with older loleaflets
-    if (success && tokens.count() > 5 && !getTokenInteger(tokens[5], "buttons", buttons))
+    if (success && tokens.size() > 5 && !getTokenInteger(tokens[5], "buttons", buttons))
         success = false;
 
     // compatibility with older loleaflets
-    if (success && tokens.count() > 6 && !getTokenInteger(tokens[6], "modifier", modifier))
+    if (success && tokens.size() > 6 && !getTokenInteger(tokens[6], "modifier", modifier))
         success = false;
 
     if (!success)
@@ -763,26 +789,26 @@ bool ChildSession::mouseEvent(const char* /*buffer*/, int /*length*/, StringToke
     return true;
 }
 
-bool ChildSession::unoCommand(const char* /*buffer*/, int /*length*/, StringTokenizer& tokens)
+bool ChildSession::unoCommand(const char* /*buffer*/, int /*length*/, const std::vector<std::string>& tokens)
 {
-    if (tokens.count() <= 1)
+    if (tokens.size() <= 1)
     {
         sendTextFrame("error: cmd=uno kind=syntax");
         return false;
     }
 
     // we need to get LOK_CALLBACK_UNO_COMMAND_RESULT callback when saving
-    const bool bNotify = (tokens[1] == ".uno:Save");
+    const bool bNotify = (tokens[1] == ".uno:Save" || tokens[1] == ".uno:Undo" || tokens[1] == ".uno:Redo");
 
     std::unique_lock<std::mutex> lock(_docManager.getDocumentMutex());
 
     getLOKitDocument()->setView(_viewId);
 
-    if (tokens.count() == 2 && tokens[1] == ".uno:fakeDiskFull")
+    if (tokens.size() == 2 && tokens[1] == ".uno:fakeDiskFull")
     {
         Util::alertAllUsers("internal", "diskfull");
     }
-    else if (tokens.count() == 2)
+    else if (tokens.size() == 2)
     {
         getLOKitDocument()->postUnoCommand(tokens[1].c_str(), nullptr, bNotify);
     }
@@ -796,10 +822,10 @@ bool ChildSession::unoCommand(const char* /*buffer*/, int /*length*/, StringToke
     return true;
 }
 
-bool ChildSession::selectText(const char* /*buffer*/, int /*length*/, StringTokenizer& tokens)
+bool ChildSession::selectText(const char* /*buffer*/, int /*length*/, const std::vector<std::string>& tokens)
 {
     int type, x, y;
-    if (tokens.count() != 4 ||
+    if (tokens.size() != 4 ||
         !getTokenKeyword(tokens[1], "type",
                          {{"start", LOK_SETTEXTSELECTION_START},
                           {"end", LOK_SETTEXTSELECTION_END},
@@ -821,10 +847,10 @@ bool ChildSession::selectText(const char* /*buffer*/, int /*length*/, StringToke
     return true;
 }
 
-bool ChildSession::selectGraphic(const char* /*buffer*/, int /*length*/, StringTokenizer& tokens)
+bool ChildSession::selectGraphic(const char* /*buffer*/, int /*length*/, const std::vector<std::string>& tokens)
 {
     int type, x, y;
-    if (tokens.count() != 4 ||
+    if (tokens.size() != 4 ||
         !getTokenKeyword(tokens[1], "type",
                          {{"start", LOK_SETGRAPHICSELECTION_START},
                           {"end", LOK_SETGRAPHICSELECTION_END}},
@@ -845,9 +871,9 @@ bool ChildSession::selectGraphic(const char* /*buffer*/, int /*length*/, StringT
     return true;
 }
 
-bool ChildSession::resetSelection(const char* /*buffer*/, int /*length*/, StringTokenizer& tokens)
+bool ChildSession::resetSelection(const char* /*buffer*/, int /*length*/, const std::vector<std::string>& tokens)
 {
-    if (tokens.count() != 1)
+    if (tokens.size() != 1)
     {
         sendTextFrame("error: cmd=resetselection kind=syntax");
         return false;
@@ -862,11 +888,11 @@ bool ChildSession::resetSelection(const char* /*buffer*/, int /*length*/, String
     return true;
 }
 
-bool ChildSession::saveAs(const char* /*buffer*/, int /*length*/, StringTokenizer& tokens)
+bool ChildSession::saveAs(const char* /*buffer*/, int /*length*/, const std::vector<std::string>& tokens)
 {
     std::string url, format, filterOptions;
 
-    if (tokens.count() < 4 ||
+    if (tokens.size() < 4 ||
         !getTokenString(tokens[1], "url", url))
     {
         sendTextFrame("error: cmd=saveas kind=syntax");
@@ -877,7 +903,7 @@ bool ChildSession::saveAs(const char* /*buffer*/, int /*length*/, StringTokenize
 
     if (getTokenString(tokens[3], "options", filterOptions))
     {
-        if (tokens.count() > 4)
+        if (tokens.size() > 4)
         {
             filterOptions += Poco::cat(std::string(" "), tokens.begin() + 4, tokens.end());
         }
@@ -903,10 +929,10 @@ bool ChildSession::saveAs(const char* /*buffer*/, int /*length*/, StringTokenize
     return true;
 }
 
-bool ChildSession::setClientPart(const char* /*buffer*/, int /*length*/, StringTokenizer& tokens)
+bool ChildSession::setClientPart(const char* /*buffer*/, int /*length*/, const std::vector<std::string>& tokens)
 {
     int part;
-    if (tokens.count() < 2 ||
+    if (tokens.size() < 2 ||
         !getTokenInteger(tokens[1], "part", part))
     {
         sendTextFrame("error: cmd=setclientpart kind=invalid");
@@ -925,10 +951,10 @@ bool ChildSession::setClientPart(const char* /*buffer*/, int /*length*/, StringT
     return true;
 }
 
-bool ChildSession::setPage(const char* /*buffer*/, int /*length*/, StringTokenizer& tokens)
+bool ChildSession::setPage(const char* /*buffer*/, int /*length*/, const std::vector<std::string>& tokens)
 {
     int page;
-    if (tokens.count() < 2 ||
+    if (tokens.size() < 2 ||
         !getTokenInteger(tokens[1], "page", page))
     {
         sendTextFrame("error: cmd=setpage kind=invalid");
@@ -943,11 +969,69 @@ bool ChildSession::setPage(const char* /*buffer*/, int /*length*/, StringTokeniz
     return true;
 }
 
-void ChildSession::loKitCallback(const int nType, const std::string& rPayload)
+/* If the user is inactive we have to remember important events so that when
+ * the user becomes active again, we can replay the events.
+ */
+void ChildSession::rememberEventsForInactiveUser(const int type, const std::string& payload)
 {
-    const auto typeName = LOKitHelper::kitCallbackTypeToString(nType);
+    if (type == LOK_CALLBACK_INVALIDATE_TILES)
+    {
+        auto lock(getLock());
+        _stateRecorder.recordInvalidate(); // TODO remember the area, not just a bool ('true' invalidates everything)
+    }
+    else if (type == LOK_CALLBACK_INVALIDATE_VISIBLE_CURSOR ||
+             type == LOK_CALLBACK_CURSOR_VISIBLE ||
+             type == LOK_CALLBACK_TEXT_SELECTION ||
+             type == LOK_CALLBACK_TEXT_SELECTION_START ||
+             type == LOK_CALLBACK_TEXT_SELECTION_END ||
+             type == LOK_CALLBACK_CELL_FORMULA ||
+             type == LOK_CALLBACK_CELL_CURSOR ||
+             type == LOK_CALLBACK_GRAPHIC_SELECTION ||
+             type == LOK_CALLBACK_DOCUMENT_SIZE_CHANGED ||
+             type == LOK_CALLBACK_INVALIDATE_HEADER ||
+             type == LOK_CALLBACK_CELL_ADDRESS)
+    {
+        auto lock(getLock());
+        _stateRecorder.recordEvent(type, payload);
+    }
+    else if (type == LOK_CALLBACK_INVALIDATE_VIEW_CURSOR ||
+             type == LOK_CALLBACK_TEXT_VIEW_SELECTION ||
+             type == LOK_CALLBACK_CELL_VIEW_CURSOR ||
+             type == LOK_CALLBACK_GRAPHIC_VIEW_SELECTION ||
+             type == LOK_CALLBACK_VIEW_CURSOR_VISIBLE ||
+             type == LOK_CALLBACK_VIEW_LOCK)
+    {
+        auto lock(getLock());
+        Poco::JSON::Parser parser;
+
+        auto root = parser.parse(payload).extract<Poco::JSON::Object::Ptr>();
+        int viewId = root->getValue<int>("viewId");
+        _stateRecorder.recordViewEvent(viewId, type, payload);
+    }
+    else if (type == LOK_CALLBACK_STATE_CHANGED)
+    {
+        std::string name;
+        std::string value;
+        if (LOOLProtocol::parseNameValuePair(payload, name, value, '='))
+        {
+            auto lock(getLock());
+            _stateRecorder.recordState(name, payload);
+        }
+    }
+    else if (type == LOK_CALLBACK_REDLINE_TABLE_SIZE_CHANGED ||
+             type == LOK_CALLBACK_REDLINE_TABLE_ENTRY_MODIFIED ||
+             type == LOK_CALLBACK_COMMENT)
+    {
+        auto lock(getLock());
+        _stateRecorder.recordEventSequence(type, payload);
+    }
+}
+
+void ChildSession::loKitCallback(const int type, const std::string& payload)
+{
+    const auto typeName = LOKitHelper::kitCallbackTypeToString(type);
     LOG_TRC("CallbackWorker::callback [" << getName() << "]: " <<
-            typeName << " [" << rPayload << "].");
+            typeName << " [" << payload << "].");
 
     if (isCloseFrame())
     {
@@ -961,54 +1045,21 @@ void ChildSession::loKitCallback(const int nType, const std::string& rPayload)
     }
     else if (!isActive())
     {
-        // Cache important notifications to replay them when our client
-        // goes inactive and loses them.
-        if (nType == LOK_CALLBACK_INVALIDATE_VISIBLE_CURSOR ||
-            nType == LOK_CALLBACK_CURSOR_VISIBLE ||
-            nType == LOK_CALLBACK_CELL_CURSOR ||
-            nType == LOK_CALLBACK_CELL_FORMULA ||
-            nType == LOK_CALLBACK_GRAPHIC_SELECTION ||
-            nType == LOK_CALLBACK_TEXT_SELECTION ||
-            nType == LOK_CALLBACK_TEXT_SELECTION_START ||
-            nType == LOK_CALLBACK_TEXT_SELECTION_END ||
-            nType == LOK_CALLBACK_DOCUMENT_SIZE_CHANGED ||
-            nType == LOK_CALLBACK_INVALIDATE_VIEW_CURSOR ||
-            nType == LOK_CALLBACK_TEXT_VIEW_SELECTION ||
-            nType == LOK_CALLBACK_CELL_VIEW_CURSOR ||
-            nType == LOK_CALLBACK_GRAPHIC_VIEW_SELECTION ||
-            nType == LOK_CALLBACK_VIEW_CURSOR_VISIBLE ||
-            nType == LOK_CALLBACK_VIEW_LOCK)
-        {
-            auto lock(getLock());
-
-            _lastDocEvents[nType] = rPayload;
-        }
-
-        if (nType == LOK_CALLBACK_STATE_CHANGED)
-        {
-            std::string name;
-            std::string value;
-            if (LOOLProtocol::parseNameValuePair(rPayload, name, value, '='))
-            {
-                auto lock(getLock());
-
-                _lastDocStates[name] = rPayload;
-            }
-        }
+        rememberEventsForInactiveUser(type, payload);
 
         // Pass save notifications through.
-        if (nType != LOK_CALLBACK_UNO_COMMAND_RESULT || rPayload.find(".uno:Save") == std::string::npos)
+        if (type != LOK_CALLBACK_UNO_COMMAND_RESULT || payload.find(".uno:Save") == std::string::npos)
         {
             LOG_TRC("Skipping callback [" << typeName << "] on inactive session " << getName());
             return;
         }
     }
 
-    switch (nType)
+    switch (type)
     {
     case LOK_CALLBACK_INVALIDATE_TILES:
         {
-            StringTokenizer tokens(rPayload, ",", StringTokenizer::TOK_IGNORE_EMPTY | StringTokenizer::TOK_TRIM);
+            StringTokenizer tokens(payload, ",", StringTokenizer::TOK_IGNORE_EMPTY | StringTokenizer::TOK_TRIM);
             if (tokens.count() == 5)
             {
                 int part, x, y, width, height;
@@ -1038,50 +1089,55 @@ void ChildSession::loKitCallback(const int nType, const std::string& rPayload)
                               " width=" + std::to_string(width) +
                               " height=" + std::to_string(height));
             }
+            else if (tokens.count() == 2 && tokens[0] == "EMPTY")
+            {
+                const std::string part = (_docType != "text" ? tokens[1].c_str() : "0"); // Writer renders everything as part 0.
+                sendTextFrame("invalidatetiles: EMPTY, " + part);
+            }
             else
             {
-                sendTextFrame("invalidatetiles: " + rPayload);
+                sendTextFrame("invalidatetiles: " + payload);
             }
         }
         break;
     case LOK_CALLBACK_INVALIDATE_VISIBLE_CURSOR:
-        sendTextFrame("invalidatecursor: " + rPayload);
+        sendTextFrame("invalidatecursor: " + payload);
         break;
     case LOK_CALLBACK_TEXT_SELECTION:
-        sendTextFrame("textselection: " + rPayload);
+        sendTextFrame("textselection: " + payload);
         break;
     case LOK_CALLBACK_TEXT_SELECTION_START:
-        sendTextFrame("textselectionstart: " + rPayload);
+        sendTextFrame("textselectionstart: " + payload);
         break;
     case LOK_CALLBACK_TEXT_SELECTION_END:
-        sendTextFrame("textselectionend: " + rPayload);
+        sendTextFrame("textselectionend: " + payload);
         break;
     case LOK_CALLBACK_CURSOR_VISIBLE:
-        sendTextFrame("cursorvisible: " + rPayload);
+        sendTextFrame("cursorvisible: " + payload);
         break;
     case LOK_CALLBACK_GRAPHIC_SELECTION:
-        sendTextFrame("graphicselection: " + rPayload);
+        sendTextFrame("graphicselection: " + payload);
         break;
     case LOK_CALLBACK_CELL_CURSOR:
-        sendTextFrame("cellcursor: " + rPayload);
+        sendTextFrame("cellcursor: " + payload);
         break;
     case LOK_CALLBACK_CELL_FORMULA:
-        sendTextFrame("cellformula: " + rPayload);
+        sendTextFrame("cellformula: " + payload);
         break;
     case LOK_CALLBACK_MOUSE_POINTER:
-        sendTextFrame("mousepointer: " + rPayload);
+        sendTextFrame("mousepointer: " + payload);
         break;
     case LOK_CALLBACK_HYPERLINK_CLICKED:
-        sendTextFrame("hyperlinkclicked: " + rPayload);
+        sendTextFrame("hyperlinkclicked: " + payload);
         break;
     case LOK_CALLBACK_STATE_CHANGED:
-        sendTextFrame("statechanged: " + rPayload);
+        sendTextFrame("statechanged: " + payload);
         break;
     case LOK_CALLBACK_SEARCH_NOT_FOUND:
-        sendTextFrame("searchnotfound: " + rPayload);
+        sendTextFrame("searchnotfound: " + payload);
         break;
     case LOK_CALLBACK_SEARCH_RESULT_SELECTION:
-        sendTextFrame("searchresultselection: " + rPayload);
+        sendTextFrame("searchresultselection: " + payload);
         break;
     case LOK_CALLBACK_DOCUMENT_SIZE_CHANGED:
         {
@@ -1102,20 +1158,19 @@ void ChildSession::loKitCallback(const int nType, const std::string& rPayload)
             lock.unlock();
 
             getStatus("", 0);
-            getPartPageRectangles("", 0);
         }
         break;
     case LOK_CALLBACK_SET_PART:
-        sendTextFrame("setpart: " + rPayload);
+        sendTextFrame("setpart: " + payload);
         break;
     case LOK_CALLBACK_UNO_COMMAND_RESULT:
-        sendTextFrame("unocommandresult: " + rPayload);
+        sendTextFrame("unocommandresult: " + payload);
         break;
     case LOK_CALLBACK_ERROR:
         {
-            LOG_ERR("CALLBACK_ERROR: " << rPayload);
+            LOG_ERR("CALLBACK_ERROR: " << payload);
             Parser parser;
-            Poco::Dynamic::Var var = parser.parse(rPayload);
+            Poco::Dynamic::Var var = parser.parse(payload);
             Object::Ptr object = var.extract<Object::Ptr>();
 
             sendTextFrame("error: cmd=" + object->get("cmd").toString() +
@@ -1123,43 +1178,52 @@ void ChildSession::loKitCallback(const int nType, const std::string& rPayload)
         }
         break;
     case LOK_CALLBACK_CONTEXT_MENU:
-        sendTextFrame("contextmenu: " + rPayload);
+        sendTextFrame("contextmenu: " + payload);
         break;
     case LOK_CALLBACK_STATUS_INDICATOR_START:
         sendTextFrame("statusindicatorstart:");
         break;
     case LOK_CALLBACK_STATUS_INDICATOR_SET_VALUE:
-        sendTextFrame("statusindicatorsetvalue: " + rPayload);
+        sendTextFrame("statusindicatorsetvalue: " + payload);
         break;
     case LOK_CALLBACK_STATUS_INDICATOR_FINISH:
         sendTextFrame("statusindicatorfinish:");
         break;
     case LOK_CALLBACK_INVALIDATE_VIEW_CURSOR:
-        sendTextFrame("invalidateviewcursor: " + rPayload);
+        sendTextFrame("invalidateviewcursor: " + payload);
         break;
     case LOK_CALLBACK_TEXT_VIEW_SELECTION:
-        sendTextFrame("textviewselection: " + rPayload);
+        sendTextFrame("textviewselection: " + payload);
         break;
     case LOK_CALLBACK_CELL_VIEW_CURSOR:
-        sendTextFrame("cellviewcursor: " + rPayload);
+        sendTextFrame("cellviewcursor: " + payload);
         break;
     case LOK_CALLBACK_GRAPHIC_VIEW_SELECTION:
-        sendTextFrame("graphicviewselection: " + rPayload);
+        sendTextFrame("graphicviewselection: " + payload);
         break;
     case LOK_CALLBACK_VIEW_CURSOR_VISIBLE:
-        sendTextFrame("viewcursorvisible: " + rPayload);
+        sendTextFrame("viewcursorvisible: " + payload);
         break;
     case LOK_CALLBACK_VIEW_LOCK:
-        sendTextFrame("viewlock: " + rPayload);
+        sendTextFrame("viewlock: " + payload);
         break;
     case LOK_CALLBACK_REDLINE_TABLE_SIZE_CHANGED:
-        sendTextFrame("redlinetablechanged: " + rPayload);
+        sendTextFrame("redlinetablechanged: " + payload);
         break;
     case LOK_CALLBACK_REDLINE_TABLE_ENTRY_MODIFIED:
-        sendTextFrame("redlinetablemodified: " + rPayload);
+        sendTextFrame("redlinetablemodified: " + payload);
+        break;
+    case LOK_CALLBACK_COMMENT:
+        sendTextFrame("comment: " + payload);
+        break;
+    case LOK_CALLBACK_INVALIDATE_HEADER:
+        sendTextFrame("invalidateheader: " + payload);
+        break;
+    case LOK_CALLBACK_CELL_ADDRESS:
+        sendTextFrame("celladdress: " + payload);
         break;
     default:
-        LOG_ERR("Unknown callback event (" << nType << "): " << rPayload);
+        LOG_ERR("Unknown callback event (" << type << "): " << payload);
     }
 }
 
