@@ -1,3 +1,4 @@
+/* -*- js-indent-level: 8 -*- */
 /*
  * L.TileLayer is used for standard xyz-numbered tile layers.
  */
@@ -48,7 +49,8 @@ L.TileLayer = L.GridLayer.extend({
 		detectRetina: true,
 		crossOrigin: false,
 		previewInvalidationTimeout: 1000,
-		defaultPermission: 'view'
+		marginX: 10,
+		marginY: 10
 	},
 
 	initialize: function (url, options) {
@@ -61,7 +63,7 @@ L.TileLayer = L.GridLayer.extend({
 		this._tileHeightPx = options.tileSize;
 
 		// detecting retina displays, adjusting tileWidthPx, tileHeightPx and zoom levels
-		if (options.detectRetina && L.Browser.retina && options.maxZoom > 0) {
+		if (options.docType !== 'spreadsheet' && options.detectRetina && L.Browser.retina && options.maxZoom > 0) {
 			this._tileWidthPx *= 2;
 			this._tileHeightPx *= 2;
 			options.zoomOffset++;
@@ -135,8 +137,6 @@ L.TileLayer = L.GridLayer.extend({
 		this._msgQueue = [];
 		this._toolbarCommandValues = {};
 		this._previewInvalidations = [];
-		this._partPageRectanglesTwips = [];
-		this._partPageRectanglesPixels = [];
 		this._clientZoom = 'tilepixelwidth=' + this._tileWidthPx + ' ' +
 			'tilepixelheight=' + this._tileHeightPx + ' ' +
 			'tiletwipwidth=' + this.options.tileWidthTwips + ' ' +
@@ -171,6 +171,63 @@ L.TileLayer = L.GridLayer.extend({
 		this._levels = {};
 		this._tiles = {};
 		this._tileCache = {};
+		this._map._socket.sendMessage('commandvalues command=.uno:ViewAnnotations');
+		var that = this;
+		$.contextMenu({
+			selector: '.loleaflet-annotation-menu',
+			trigger: 'none',
+			className: 'loleaflet-font',
+			items: {
+				modify: {
+					name: _('Modify'),
+					callback: function (key, options) {
+						that.onAnnotationModify.call(that, options.$trigger.get(0).annotation);
+					}
+				},
+				reply: (this._docType !== 'text') ? undefined : {
+					name: _('Reply'),
+					callback: function (key, options) {
+						that.onAnnotationReply.call(that, options.$trigger.get(0).annotation);
+					}
+				},
+				remove: {
+					name: _('Remove'),
+					callback: function (key, options) {
+						that.onAnnotationRemove.call(that, options.$trigger.get(0).annotation._data.id);
+					}
+				}
+			},
+			events: {
+				show: function (options) {
+					options.$trigger.get(0).annotation._contextMenu = true;
+				},
+				hide: function (options) {
+					options.$trigger.get(0).annotation._contextMenu = false;
+				}
+			}
+		});
+		$.contextMenu({
+			selector: '.loleaflet-annotation-menu-redline',
+			trigger: 'none',
+			className: 'loleaflet-font',
+			items: {
+				modify: {
+					name: _('Comment'),
+					callback: function (key, options) {
+						that.onAnnotationModify.call(that, options.$trigger.get(0).annotation);
+					}
+				}
+			},
+			events: {
+				show: function (options) {
+					options.$trigger.get(0).annotation._contextMenu = true;
+				},
+				hide: function (options) {
+					options.$trigger.get(0).annotation._contextMenu = false;
+				}
+			}
+		});
+		this._map._socket.sendMessage('commandvalues command=.uno:AcceptTrackedChanges');
 
 		map._fadeAnimated = false;
 		this._viewReset();
@@ -183,7 +240,6 @@ L.TileLayer = L.GridLayer.extend({
 		map.on('drop', this._onDrop, this);
 
 		map.on('zoomend', this._onUpdateCursor, this);
-		map.on('zoomend', this._onUpdatePartPageRectangles, this);
 		if (this._docType === 'spreadsheet') {
 			map.on('zoomend', this._onCellCursorShift, this);
 		}
@@ -308,6 +364,9 @@ L.TileLayer = L.GridLayer.extend({
 		else if (textMsg.startsWith('cellcursor:')) {
 			this._onCellCursorMsg(textMsg);
 		}
+		else if (textMsg.startsWith('celladdress:')) {
+			this._onCellAddressMsg(textMsg);
+		}
 		else if (textMsg.startsWith('cellformula:')) {
 			this._onCellFormulaMsg(textMsg);
 		}
@@ -339,9 +398,6 @@ L.TileLayer = L.GridLayer.extend({
 		else if (textMsg.startsWith('mousepointer:')) {
 			this._onMousePointerMsg(textMsg);
 		}
-		else if (textMsg.startsWith('partpagerectangles:')) {
-			this._onPartPageRectanglesMsg(textMsg);
-		}
 		else if (textMsg.startsWith('renderfont:')) {
 			this._onRenderFontMsg(textMsg, img);
 		}
@@ -430,13 +486,15 @@ L.TileLayer = L.GridLayer.extend({
 		}
 		else if (obj.commandName === '.uno:CellCursor') {
 			this._onCellCursorMsg(obj.commandValues);
-		} else if (this._map.unoToolbarCommands.indexOf(obj.commandName) !== -1) {
+		}
+		else if (this._map.unoToolbarCommands.indexOf(obj.commandName) !== -1) {
 			this._toolbarCommandValues[obj.commandName] = obj.commandValues;
 			this._map.fire('updatetoolbarcommandvalues', {
 				commandName: obj.commandName,
 				commandValues: obj.commandValues
 			});
-		} else {
+		}
+		else {
 			this._map.fire('commandvalues', {
 				commandName: obj.commandName,
 				commandValues: obj.commandValues
@@ -444,6 +502,11 @@ L.TileLayer = L.GridLayer.extend({
 		}
 	},
 
+	_onCellAddressMsg: function (textMsg) {
+		var address = textMsg.substring(13);
+		this._map.fire('celladdress', {address: address});
+	},
+
 	_onCellFormulaMsg: function (textMsg) {
 		var formula = textMsg.substring(13);
 		this._selectionTextContent = formula;
@@ -477,7 +540,7 @@ L.TileLayer = L.GridLayer.extend({
 		else if (command.id === 'slideshow') {
 			this._map.fire('slidedownloadready', {url: url});
 		}
-		else {
+		else if (command.id === 'export') {
 			this._map._fileDownloader.src = url;
 		}
 	},
@@ -787,19 +850,19 @@ L.TileLayer = L.GridLayer.extend({
 		this._onUpdateViewCursor(viewId);
 	},
 
-	_addView: function(viewId, userid, username, color) {
-		if (color === 0 && this._map.getDocType() !== 'text') {
-			color = L.LOUtil.getViewIdColor(viewId);
+	_addView: function(viewInfo) {
+		if (viewInfo.color === 0 && this._map.getDocType() !== 'text') {
+			viewInfo.color = L.LOUtil.getViewIdColor(viewInfo.id);
 		}
 
-		this._map.addView(viewId, userid, username, color);
+		this._map.addView(viewInfo);
 
 		//TODO: We can initialize color and other properties here.
-		if (typeof this._viewCursors[viewId] !== 'undefined') {
-			this._viewCursors[viewId] = {};
+		if (typeof this._viewCursors[viewInfo.id] !== 'undefined') {
+			this._viewCursors[viewInfo.id] = {};
 		}
 
-		this._onUpdateViewCursor(viewId);
+		this._onUpdateViewCursor(viewInfo.id);
 	},
 
 	_removeView: function(viewId) {
@@ -833,7 +896,7 @@ L.TileLayer = L.GridLayer.extend({
 		var viewIds = [];
 		for (var viewInfoIdx in viewInfo) {
 			if (!(parseInt(viewInfo[viewInfoIdx].id) in this._map._viewInfo)) {
-				this._addView(viewInfo[viewInfoIdx].id, viewInfo[viewInfoIdx].userid, viewInfo[viewInfoIdx].username, viewInfo[viewInfoIdx].color);
+				this._addView(viewInfo[viewInfoIdx]);
 			}
 			viewIds.push(viewInfo[viewInfoIdx].id);
 		}
@@ -846,49 +909,24 @@ L.TileLayer = L.GridLayer.extend({
 		}
 	},
 
-	_onPartPageRectanglesMsg: function (textMsg) {
-		textMsg = textMsg.substring(19);
-		var pages = textMsg.split(';');
-		this._partPageRectanglesTwips = [];
-		this._partPageRectanglesPixels = [];
-		for (var i = 0; i < pages.length; i++) {
-			var strTwips = pages[i].match(/\d+/g);
-			if (!strTwips) {
-				// probably not a text file
-				return;
-			}
-			var topLeftTwips = new L.Point(parseInt(strTwips[0]), parseInt(strTwips[1]));
-			var offset = new L.Point(parseInt(strTwips[2]), parseInt(strTwips[3]));
-			var bottomRightTwips = topLeftTwips.add(offset);
-			var pageBoundsTwips = new L.Bounds(topLeftTwips, bottomRightTwips);
-			this._partPageRectanglesTwips.push(pageBoundsTwips);
-			var pageBoundsPixels = new L.Bounds(
-					this._twipsToPixels(topLeftTwips),
-					this._twipsToPixels(bottomRightTwips));
-			this._partPageRectanglesPixels.push(pageBoundsPixels);
-		}
-		this._map.fire('partpagerectangles', {
-			pixelRectangles: this._partPageRectanglesPixels,
-			twipsRectangles: this._partPageRectanglesTwips
-		});
-		this._onCurrentPageUpdate();
-	},
-
 	_onRenderFontMsg: function (textMsg, img) {
 		var command = this._map._socket.parseServerCmd(textMsg);
 		this._map.fire('renderfont', {
 			font: command.font,
+			char: command.char,
 			img: img
 		});
 	},
 
 	_onSearchNotFoundMsg: function (textMsg) {
 		this._clearSearchResults();
+		this._searchRequested = false;
 		var originalPhrase = textMsg.substring(16);
 		this._map.fire('search', {originalPhrase: originalPhrase, count: 0});
 	},
 
 	_onSearchResultSelection: function (textMsg) {
+		this._searchRequested = false;
 		textMsg = textMsg.substring(23);
 		var obj = JSON.parse(textMsg);
 		var originalPhrase = obj.searchString;
@@ -909,12 +947,15 @@ L.TileLayer = L.GridLayer.extend({
 			this._clearSearchResults();
 			this._searchResults = results;
 			this._map.setPart(results[0].part); // go to first result.
+		} else if (count === 1) {
+			this._lastSearchResult = results[0];
 		}
 		this._searchTerm = originalPhrase;
 		this._map.fire('search', {originalPhrase: originalPhrase, count: count, highlightAll: highlightAll, results: results});
 	},
 
 	_clearSearchResults: function() {
+		this._lastSearchResult = null;
 		this._searchResults = null;
 		this._searchTerm = null;
 		this._searchResultsLayer.clearLayers();
@@ -974,8 +1015,7 @@ L.TileLayer = L.GridLayer.extend({
 		}
 
 		this._map.hideBusy();
-		// TODO when implemented in the LOK, add also obj.result
-		this._map.fire('commandresult', {commandName: commandName, success: success});
+		this._map.fire('commandresult', {commandName: commandName, success: success, result: obj.result});
 
 	},
 
@@ -1135,6 +1175,9 @@ L.TileLayer = L.GridLayer.extend({
 			});
 		}
 		else if (tile) {
+			if (command.hash != undefined) {
+				tile.oldhash = command.hash;
+			}
 			if (this._tiles[key]._invalidCount > 0) {
 				this._tiles[key]._invalidCount -= 1;
 			}
@@ -1146,9 +1189,6 @@ L.TileLayer = L.GridLayer.extend({
 			}
 			tile.el.src = img;
 		}
-		else if (command.preFetch === 'true') {
-			this._tileCache[key] = img;
-		}
 		L.Log.log(textMsg, L.INCOMING, key);
 	},
 
@@ -1219,8 +1259,7 @@ L.TileLayer = L.GridLayer.extend({
 	},
 
 	_postKeyboardEvent: function(type, charcode, keycode) {
-		if (this._docType === 'spreadsheet' && this._prevCellCursor && type === 'input')
-		{
+		if (this._docType === 'spreadsheet' && this._prevCellCursor && type === 'input') {
 			if (keycode === 1030) { // PgUp
 				if (this._cellCursorOnPgUp) {
 					return;
@@ -1273,8 +1312,6 @@ L.TileLayer = L.GridLayer.extend({
 
 	// Update cursor layer (blinking cursor).
 	_onUpdateCursor: function (e) {
-		var pixBounds = L.bounds(this._map.latLngToLayerPoint(this._visibleCursor.getSouthWest()),
-						 this._map.latLngToLayerPoint(this._visibleCursor.getNorthEast()));
 		var cursorPos = this._visibleCursor.getNorthWest();
 
 		if (!e && !this._map.getBounds().contains(this._visibleCursor) && this._isCursorVisible) {
@@ -1340,7 +1377,7 @@ L.TileLayer = L.GridLayer.extend({
 		var viewCursorVisible = this._viewCursors[viewId].visible;
 		var viewPart = this._viewCursors[viewId].part;
 
-		if (viewCursorVisible && !this._isEmptyRectangle(this._viewCursors[viewId].bounds) &&
+		if (!this._map.isViewReadOnly(viewId) && viewCursorVisible && !this._isEmptyRectangle(this._viewCursors[viewId].bounds) &&
 		   (this._docType === 'text' || this._selectedPart === viewPart)) {
 			if (!viewCursorMarker) {
 				var viewCursorOptions = {
@@ -1564,17 +1601,18 @@ L.TileLayer = L.GridLayer.extend({
 						scrollY = (this._cellCursor.getNorth() - mapBounds.getNorth()) + mapHalfHeight;
 					}
 				}
-				else {
+				else if (horizontalDirection !== 0 || verticalDirection != 0) {
 					var spacingX = Math.abs(this._cellCursor.getEast() - this._cellCursor.getWest()) / 4.0;
 					var spacingY = Math.abs((this._cellCursor.getSouth() - this._cellCursor.getNorth())) / 4.0;
-					if (horizontalDirection === -1 && this._cellCursor.getWest() < mapBounds.getWest()) {
+
+					if (this._cellCursor.getWest() < mapBounds.getWest()) {
 						scrollX = this._cellCursor.getWest() - mapBounds.getWest() - spacingX;
-					} else if (horizontalDirection === 1 && this._cellCursor.getEast() > mapBounds.getEast()) {
+					} else if (this._cellCursor.getEast() > mapBounds.getEast()) {
 						scrollX = this._cellCursor.getEast() - mapBounds.getEast() + spacingX;
 					}
-					if (verticalDirection === 1 && this._cellCursor.getNorth() > mapBounds.getNorth()) {
+					if (this._cellCursor.getNorth() > mapBounds.getNorth()) {
 						scrollY = this._cellCursor.getNorth() - mapBounds.getNorth() + spacingY;
-					} else if (verticalDirection === -1 && this._cellCursor.getSouth() < mapBounds.getSouth()) {
+					} else if (this._cellCursor.getSouth() < mapBounds.getSouth()) {
 						scrollY = this._cellCursor.getSouth() - mapBounds.getSouth() - spacingY;
 					}
 				}
@@ -1689,8 +1727,9 @@ L.TileLayer = L.GridLayer.extend({
 
 			// remember the copied text, for rich copy/paste inside a document
 			this._selectionTextHash = this._selectionTextContent;
-			this._map._socket.sendMessage('uno .uno:Copy');
 		}
+
+		this._map._socket.sendMessage('uno .uno:Copy');
 	},
 
 	_onCut: function (e) {
@@ -1701,23 +1740,24 @@ L.TileLayer = L.GridLayer.extend({
 
 			// remember the copied text, for rich copy/paste inside a document
 			this._selectionTextHash = this._selectionTextContent;
-			this._map._socket.sendMessage('uno .uno:Cut');
 		}
+
+		this._map._socket.sendMessage('uno .uno:Cut');
 	},
 
 	_onPaste: function (e) {
 		e = e.originalEvent;
 		e.preventDefault();
 		var pasteString = L.Compatibility.clipboardGet(e);
-		if (pasteString) {
-			if (pasteString === this._selectionTextHash) {
-				// content of the clipboard did not change, we can do rich
-				// paste
-				this._map._socket.sendMessage('uno .uno:Paste');
-			}
-			else {
-				this._map._socket.sendMessage('paste mimetype=text/plain;charset=utf-8\n' + pasteString);
-			}
+		if (pasteString === 'false' || !pasteString || pasteString === this._selectionTextHash) {
+			// If there is nothing to paste in clipboard, no harm in
+			// issuing a .uno:Paste in case there is something internally copied in the document
+			// or if the content of the clipboard did not change, we surely must do a rich paste
+			// instead of a normal paste
+			this._map._socket.sendMessage('uno .uno:Paste');
+		}
+		else {
+			this._map._socket.sendMessage('paste mimetype=text/plain;charset=utf-8\n' + pasteString);
 		}
 	},
 
@@ -1794,7 +1834,7 @@ L.TileLayer = L.GridLayer.extend({
 	_fitWidthZoom: function (e, maxZoom) {
 		var size = e ? e.newSize : this._map.getSize();
 		var widthTwips = size.x * this._map.options.tileWidthTwips / this._tileSize;
-		maxZoom = maxZoom ? maxZoom : this._map.options.zoom;
+		maxZoom = maxZoom ? maxZoom : this._map.getZoom();
 
 		// 'fit width zoom' has no use in spreadsheets, ignore it there
 		if (this._docType !== 'spreadsheet') {
@@ -1802,7 +1842,7 @@ L.TileLayer = L.GridLayer.extend({
 			if (this._docWidthTwips > 0)
 			{
 				var ratio = widthTwips / this._docWidthTwips;
-				var zoom = this._map.options.zoom + Math.floor(Math.log(ratio) / Math.log(crsScale));
+				var zoom = this._map.getZoom() + Math.floor(Math.log(ratio) / Math.log(crsScale));
 
 				zoom = Math.max(1, zoom);
 				zoom = Math.min(maxZoom, zoom);
@@ -1831,22 +1871,6 @@ L.TileLayer = L.GridLayer.extend({
 		}
 	},
 
-	_onUpdatePartPageRectangles: function () {
-		if (this._partPageRectanglesPixels.length > 0) {
-			this._partPageRectanglesPixels = [];
-			for (var i = 0; i < this._partPageRectanglesTwips.length; i++) {
-				var pageBounds = new L.Bounds(
-						this._twipsToPixels(this._partPageRectanglesTwips[i].min),
-						this._twipsToPixels(this._partPageRectanglesTwips[i].max));
-				this._partPageRectanglesPixels.push(pageBounds);
-			}
-			this._map.fire('partpagerectangles', {
-				pixelRectangles: this._partPageRectanglesPixels,
-				twipsRectangles: this._partPageRectanglesTwips
-			});
-		}
-	},
-
 	// Cells can change position during changes of zoom level in calc
 	// hence we need to request an updated cell cursor position for this level.
 	_onCellCursorShift: function (force) {
@@ -1970,18 +1994,22 @@ L.TileLayer = L.GridLayer.extend({
 			}
 			this._debugInfo = new L.LayerGroup();
 			this._debugInfo2 = new L.LayerGroup();
+			this._debugAlwaysActive = new L.LayerGroup();
 			this._debugTyper = new L.LayerGroup();
 			map.addLayer(this._debugInfo);
 			map.addLayer(this._debugInfo2);
 			var overlayMaps = {
 				'Tile overlays': this._debugInfo,
 				'Screen overlays': this._debugInfo2,
+				'Always active': this._debugAlwaysActive,
 				'Typing': this._debugTyper
 			};
 			L.control.layers({}, overlayMaps, {collapsed: false}).addTo(map);
 
 			this._map.on('layeradd', function(e) {
-				if (e.layer === this._debugTyper) {
+				if (e.layer === this._debugAlwaysActive) {
+					map._debugAlwaysActive = true;
+				} else if (e.layer === this._debugTyper) {
 					this._debugTypeTimeout();
 				} else if (e.layer === this._debugInfo2) {
 					for (var i = 0; i < this._debugDataNames.length; i++) {
@@ -1990,7 +2018,9 @@ L.TileLayer = L.GridLayer.extend({
 				}
 			}, this);
 			map.on('layerremove', function(e) {
-				if (e.layer === this._debugTyper) {
+				if (e.layer === this._debugAlwaysActive) {
+					map._debugAlwaysActive = false;
+				} else if (e.layer === this._debugTyper) {
 					clearTimeout(this._debugTypeTimeoutId);
 				} else if (e.layer === this._debugInfo2) {
 					for (var i in this._debugData) {
@@ -2047,6 +2077,7 @@ L.TileLayer = L.GridLayer.extend({
 		if (tile._debugTile) {
 			tile._debugTile.setStyle({fillOpacity: 0.5, fillColor: 'blue'});
 			tile._debugTime.date = +new Date();
+			tile._debugTile.date = +new Date();
 			tile._debugInvalidateCount++;
 			this._debugInvalidateCount++;
 		}
@@ -2081,6 +2112,14 @@ L.TileLayer = L.GridLayer.extend({
 					rect.setStyle({fillOpacity: opac - 0.04});
 				}
 			}
+			for (var key in this._debugTiles) {
+				var rect = this._debugTiles[key];
+				var col = rect.options.fillColor;
+				var opac = rect.options.fillOpacity;
+				if (col === 'blue' && opac >= 0.04 && rect.date + 1000 < +new Date()) {
+					rect.setStyle({fillOpacity: opac - 0.04});
+				}
+			}
 			this._debugTimeoutId = setTimeout(function () { map._docLayer._debugTimeout(); }, 50);
 		}
 	},
