@@ -7,43 +7,30 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  */
 
-#include "config.h"
+#ifndef INCLUDED_TEST_HELPERS_HPP
+#define INCLUDED_TEST_HELPERS_HPP
 
-#include <algorithm>
-#include <condition_variable>
-#include <cstdlib>
-#include <mutex>
-#include <regex>
-#include <thread>
+#include <iterator>
 
 #include <Poco/BinaryReader.h>
-#include <Poco/DirectoryIterator.h>
 #include <Poco/Dynamic/Var.h>
-#include <Poco/FileStream.h>
 #include <Poco/JSON/JSON.h>
 #include <Poco/JSON/Parser.h>
-#include <Poco/Net/AcceptCertificateHandler.h>
 #include <Poco/Net/HTTPClientSession.h>
 #include <Poco/Net/HTTPRequest.h>
 #include <Poco/Net/HTTPResponse.h>
 #include <Poco/Net/HTTPSClientSession.h>
-#include <Poco/Net/InvalidCertificateHandler.h>
 #include <Poco/Net/NetException.h>
-#include <Poco/Net/PrivateKeyPassphraseHandler.h>
-#include <Poco/Net/SSLManager.h>
 #include <Poco/Net/Socket.h>
 #include <Poco/Path.h>
-#include <Poco/StreamCopier.h>
 #include <Poco/StringTokenizer.h>
-#include <Poco/Thread.h>
 #include <Poco/URI.h>
+
 #include <cppunit/extensions/HelperMacros.h>
 
 #include <Common.hpp>
 #include "common/FileUtil.hpp"
-#include <Protocol.hpp>
 #include <LOOLWebSocket.hpp>
-#include <UserMessages.hpp>
 #include <Util.hpp>
 
 #ifndef TDOC
@@ -112,9 +99,10 @@ std::vector<char> readDataFromFile(std::unique_ptr<std::fstream>& file)
 }
 
 inline
-void getDocumentPathAndURL(const std::string& docFilename, std::string& documentPath, std::string& documentURL)
+void getDocumentPathAndURL(const std::string& docFilename, std::string& documentPath, std::string& documentURL, std::string prefix)
 {
-    documentPath = FileUtil::getTempFilePath(TDOC, docFilename);
+    std::replace(prefix.begin(), prefix.end(), ' ', '_');
+    documentPath = FileUtil::getTempFilePath(TDOC, docFilename, prefix);
     std::string encodedUri;
     Poco::URI::encode("file://" + Poco::Path(documentPath).makeAbsolute().toString(), ":/?", encodedUri);
     documentURL = "lool/" + encodedUri + "/ws";
@@ -124,7 +112,11 @@ void getDocumentPathAndURL(const std::string& docFilename, std::string& document
 inline
 void sendTextFrame(LOOLWebSocket& socket, const std::string& string, const std::string& name = "")
 {
+#ifndef FUZZER
     std::cerr << name << "Sending " << string.size() << " bytes: " << LOOLProtocol::getAbbreviatedMessage(string) << std::endl;
+#else
+    (void) name;
+#endif
     socket.sendFrame(string.data(), string.size());
 }
 
@@ -145,7 +137,7 @@ Poco::Net::HTTPClientSession* createSession(const Poco::URI& uri)
 }
 
 inline
-std::string getTestServerURI()
+std::string const & getTestServerURI()
 {
     static const char* clientPort = std::getenv("LOOL_TEST_CLIENT_PORT");
 
@@ -161,7 +153,7 @@ std::string getTestServerURI()
 }
 
 inline
-int getErrorCode(LOOLWebSocket& ws, std::string& message)
+int getErrorCode(LOOLWebSocket& ws, std::string& message, const std::string& testname = "")
 {
     int flags = 0;
     int bytes = 0;
@@ -169,13 +161,13 @@ int getErrorCode(LOOLWebSocket& ws, std::string& message)
     Poco::Buffer<char> buffer(READ_BUFFER_SIZE);
 
     message.clear();
-    Poco::Timespan timeout(5000000);
-    ws.setReceiveTimeout(timeout);
+    ws.setReceiveTimeout(Poco::Timespan(5000000));
     do
     {
         bytes = ws.receiveFrame(buffer.begin(), READ_BUFFER_SIZE, flags);
+        std::cerr << testname << "Got " << LOOLProtocol::getAbbreviatedFrameDump(buffer.begin(), bytes, flags) << std::endl;
     }
-    while ((flags & Poco::Net::WebSocket::FRAME_OP_BITMASK) != Poco::Net::WebSocket::FRAME_OP_CLOSE);
+    while (bytes > 0 && (flags & Poco::Net::WebSocket::FRAME_OP_BITMASK) != Poco::Net::WebSocket::FRAME_OP_CLOSE);
 
     if (bytes > 0)
     {
@@ -187,6 +179,12 @@ int getErrorCode(LOOLWebSocket& ws, std::string& message)
     return statusCode;
 }
 
+inline
+int getErrorCode(const std::shared_ptr<LOOLWebSocket>& ws, std::string& message, const std::string& testname = "")
+{
+    return getErrorCode(*ws, message, testname);
+}
+
 inline
 std::vector<char> getResponseMessage(LOOLWebSocket& ws, const std::string& prefix, std::string name = "", const size_t timeoutMs = 10000)
 {
@@ -210,9 +208,9 @@ std::vector<char> getResponseMessage(LOOLWebSocket& ws, const std::string& prefi
                     timedout = false;
                 }
 
-                response.resize(READ_BUFFER_SIZE);
+                response.resize(READ_BUFFER_SIZE * 8);
                 int bytes = ws.receiveFrame(response.data(), response.size(), flags);
-                response.resize(bytes >= 0 ? bytes : 0);
+                response.resize(std::max(bytes, 0));
                 std::cerr << name << "Got " << LOOLProtocol::getAbbreviatedFrameDump(response.data(), bytes, flags) << std::endl;
                 const auto message = LOOLProtocol::getFirstLine(response);
                 if (bytes > 0 && (flags & Poco::Net::WebSocket::FRAME_OP_BITMASK) != Poco::Net::WebSocket::FRAME_OP_CLOSE)
@@ -221,22 +219,6 @@ std::vector<char> getResponseMessage(LOOLWebSocket& ws, const std::string& prefi
                     {
                         return response;
                     }
-                    else if (LOOLProtocol::matchPrefix("nextmessage", message))
-                    {
-                        int size = 0;
-                        if (LOOLProtocol::getTokenIntegerFromMessage(message, "size", size) && size > 0)
-                        {
-                            response.resize(size);
-                            bytes = ws.receiveFrame(response.data(), response.size(), flags);
-                            response.resize(bytes >= 0 ? bytes : 0);
-                            std::cerr << name << "Got " << LOOLProtocol::getAbbreviatedFrameDump(response.data(), bytes, flags) << std::endl;
-                            if (bytes > 0 &&
-                                LOOLProtocol::matchPrefix(prefix, LOOLProtocol::getFirstLine(response)))
-                            {
-                                return response;
-                            }
-                        }
-                    }
                 }
                 else
                 {
@@ -315,7 +297,7 @@ template <typename T>
 std::string assertNotInResponse(T& ws, const std::string& prefix, const std::string name = "")
 {
     const auto res = getResponseString(ws, prefix, name, 1000);
-    CPPUNIT_ASSERT_MESSAGE("Did not expect getting message [" + res + "].", res.empty());
+    CPPUNIT_ASSERT_MESSAGE(name + "Did not expect getting message [" + res + "].", res.empty());
     return res;
 }
 
@@ -324,7 +306,7 @@ bool isDocumentLoaded(LOOLWebSocket& ws, const std::string& name = "", bool isVi
 {
     const std::string prefix = isView ? "status:" : "statusindicatorfinish:";
     const auto message = getResponseString(ws, prefix, name);
-    return LOOLProtocol::getFirstToken(message) == prefix;
+    return LOOLProtocol::matchPrefix(prefix, message);
 }
 
 inline
@@ -404,7 +386,7 @@ std::shared_ptr<LOOLWebSocket> loadDocAndGetSocket(const std::string& docFilenam
     try
     {
         std::string documentPath, documentURL;
-        getDocumentPathAndURL(docFilename, documentPath, documentURL);
+        getDocumentPathAndURL(docFilename, documentPath, documentURL, name);
         return loadDocAndGetSocket(uri, documentURL, name, isView);
     }
     catch (const Poco::Exception& exc)
@@ -561,6 +543,63 @@ inline void sendText(std::shared_ptr<LOOLWebSocket>& socket, const std::string&
     }
 }
 
+inline std::vector<char> getTileAndSave(std::shared_ptr<LOOLWebSocket>& socket,
+                                        const std::string& req,
+                                        const std::string& filename,
+                                        const std::string& testname)
+{
+    std::cerr << testname << "Requesting: " << req << std::endl;
+    sendTextFrame(socket, req, testname);
+
+    const auto tile = getResponseMessage(socket, "tile:", testname);
+    std::cerr << testname << " Tile PNG size: " << tile.size() << std::endl;
+
+    const std::string firstLine = LOOLProtocol::getFirstLine(tile);
+    std::vector<char> res(tile.begin() + firstLine.size() + 1, tile.end());
+    std::stringstream streamRes;
+    std::copy(res.begin(), res.end(), std::ostream_iterator<char>(streamRes));
+
+    if (!filename.empty())
+    {
+        std::fstream outStream(filename, std::ios::out);
+        outStream.write(res.data(), res.size());
+        outStream.close();
+        std::cerr << testname << "Saved [" << firstLine << "] to [" << filename << "]" << std::endl;
+    }
+
+    return res;
 }
 
+inline void getServerVersion(LOOLWebSocket& socket,
+                             int& major, int& minor,
+                             const std::string& testname)
+{
+    const std::string clientVersion = "loolclient 0.1";
+    sendTextFrame(socket, clientVersion);
+    std::vector<char> loVersion = getResponseMessage(socket, "lokitversion", testname);
+    std::string line = LOOLProtocol::getFirstLine(loVersion.data(), loVersion.size());
+    line = line.substr(strlen("lokitversion "));
+    Poco::JSON::Parser parser;
+    Poco::Dynamic::Var loVersionVar = parser.parse(line);
+    const Poco::SharedPtr<Poco::JSON::Object>& loVersionObject = loVersionVar.extract<Poco::JSON::Object::Ptr>();
+    std::string loProductVersion = loVersionObject->get("ProductVersion").toString();
+    std::istringstream stream(loProductVersion);
+    stream >> major;
+    assert(stream.get() == '.');
+    stream >> minor;
+
+    std::cerr << testname << "Client [" << major << '.' << minor << "]." << std::endl;
+}
+
+inline void getServerVersion(std::shared_ptr<LOOLWebSocket>& socket,
+                             int& major, int& minor,
+                             const std::string& testname)
+{
+    getServerVersion(*socket, major, minor, testname);
+}
+
+}
+
+#endif
+
 /* vim:set shiftwidth=4 softtabstop=4 expandtab: */
